/*
 * linux/drivers/video/ep93xxfb.c -- Frame buffer device for the EP93xx
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/reboot.h>
#include <video/ep93xxfb.h>
#include <asm/arch/cx25871.h>
#include <asm/uaccess.h>

dma_addr_t fb_phys;
#define devicename				"ep93xxfb"

static  int
ep93xxfb_ioctl(struct inode *inode, struct file *file,
               unsigned int cmd, unsigned long arg, struct fb_info *info);

static unsigned long
isqrt(unsigned long a)
{
        unsigned long rem = 0;
        unsigned long root = 0;
        int i;
                                                                                                                             
        for (i = 0; i < 16; i++) {
                root <<= 1;
                rem = ((rem << 2) + (a >> 30));
                a <<= 2;
                root++;
                if (root <= rem) {
                        rem -= root;
                        root++;
                } else
                        root--;
        }
        return root >> 1;
}



struct ep93xxfb_videotimings
{
	const char *name[7];
	void (*configure)(struct fb_info *info);
	void (*on)(struct fb_info *info);
	void (*off)(struct fb_info *info);

	unsigned long hres;
	unsigned long hfp;
	unsigned long hsync;
	unsigned long hbp;

	unsigned long vres;
	unsigned long vfp;
	unsigned long vsync;
	unsigned long vbp;

	unsigned long refresh;
	unsigned long mrefresh;
	unsigned long width;
	unsigned long height;

	unsigned long clk_src:1;
	unsigned long clk_edge:1;
	unsigned long pol_blank:1;
	unsigned long pol_hsync:1;
	unsigned long pol_vsync:1;
};

#define POL_HIGH				1
#define POL_LOW					0
#define EDGE_RISING				1
#define EDGE_FALLING				0
#define CLK_INTERNAL				1
#define CLK_EXTERNAL				0

#define GPIO_DR					GPIO_PGDR
#define GPIO_DDR				GPIO_PGDDR
#define GPIO_CLK				(1 << 0)
#define GPIO_DAT				(1 << 1)
#define DELAY_USEC				2
#define READ_TIMEOUT				100


#define EE_DELAY_USEC			2
#define EE_READ_TIMEOUT			100
#define CX25871_DEV_ADDRESS		0x88


#define GPIOG_EEDAT 2
#define GPIOG_EECLK 1

static int
write_reg(unsigned char ucRegAddr, unsigned char ucRegValue)
{
  	unsigned long uiVal, uiDDR;
	unsigned char ucData, ucIdx, ucBit;
	unsigned long ulTimeout;

	uiVal = inl(GPIO_PGDR);
	uiDDR = inl(GPIO_PGDDR);

	uiVal |= (GPIOG_EEDAT | GPIOG_EECLK);
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiDDR |= (GPIOG_EEDAT | GPIOG_EECLK);
	outl( uiDDR, GPIO_PGDDR );

	udelay( EE_DELAY_USEC );

	uiVal &= ~GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal &= ~GPIOG_EECLK;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	for (ucIdx = 0; ucIdx < 3; ucIdx++) {
		if (ucIdx == 0)
			ucData = (unsigned char)CX25871_DEV_ADDRESS;
		else if (ucIdx == 1)
			ucData = ucRegAddr;
		else
			ucData = ucRegValue;

		for (ucBit = 0; ucBit < 8; ucBit++) {
			if (ucData & 0x80)
				uiVal |= GPIOG_EEDAT;
			else
				uiVal &= ~GPIOG_EEDAT;
			outl( uiVal, GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			outl( (uiVal | GPIOG_EECLK), GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			outl( uiVal, GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			ucData <<= 1;
		}

		uiDDR &= ~GPIOG_EEDAT;
		outl( uiDDR, GPIO_PGDDR );

		udelay( EE_DELAY_USEC );

		outl( (uiVal | GPIOG_EECLK), GPIO_PGDR );

		udelay( EE_DELAY_USEC );

		ulTimeout = 0;
		while ( inl(GPIO_PGDR) & GPIOG_EEDAT ) {
			udelay( EE_DELAY_USEC );
			ulTimeout++;
			if (ulTimeout > EE_READ_TIMEOUT )
				return 1;
		}

		outl( uiVal, GPIO_PGDR );

		udelay( EE_DELAY_USEC );

		uiDDR |= GPIOG_EEDAT;
		outl( uiDDR, GPIO_PGDDR );
		udelay( EE_DELAY_USEC );
	}

	uiVal &= ~GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal |= GPIOG_EECLK;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal |= GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );


	return 0;
}

static void
philips_lb064v02_on(struct fb_info *info)
{
	outl(inl(GPIO_PADDR) | 2, GPIO_PADDR);
	outl(inl(GPIO_PADR) | 2, GPIO_PADR);
}

static void
philips_lb064v02_off(struct fb_info *info)
{
	outl(inl(GPIO_PADR) & ~2, GPIO_PADR);
}

static int
conexant_cx25871_reboot(struct notifier_block *this, unsigned long event,
			void *x)
{
	switch (event) {
	case SYS_HALT:
	case SYS_POWER_OFF:
	case SYS_RESTART:
		write_reg(0xba, 0x80); //0x88
	default:
		return NOTIFY_DONE;
	}
}

static struct notifier_block conexant_cx25871_notifier = {
	conexant_cx25871_reboot,
	NULL,
	5
};

static void
conexant_cx25871_configure_ntsc(struct fb_info *info)
{
	write_reg(0xb8, 0x00);
       	write_reg(0xBA, CX25871_REGxBA_SLAVER | CX25871_REGxBA_DACOFF);     
	write_reg(0xC6, (CX25871_REGxC6_INMODE_MASK & 0x3));
	write_reg(0xC4, CX25871_REGxC4_EN_OUT);
	write_reg(0x32, 0);
	write_reg(0xBA, CX25871_REGxBA_SLAVER );

	register_reboot_notifier(&conexant_cx25871_notifier);
}

static void
conexant_cx25871_configure_pal(struct fb_info *info)
{
	write_reg(0xb8, 0x01);
       	write_reg(0xBA, CX25871_REGxBA_SLAVER | CX25871_REGxBA_DACOFF);     
	write_reg(0xC6, (CX25871_REGxC6_INMODE_MASK & 0x3));
	write_reg(0xC4, CX25871_REGxC4_EN_OUT);
	write_reg(0x32, 0);
	write_reg(0xBA, CX25871_REGxBA_SLAVER );

	register_reboot_notifier(&conexant_cx25871_notifier);
}



static void
conexant_cx25871_overscan_configure(struct fb_info *info)
{
	write_reg(0xb8, 0x00);
	write_reg(0xba, 0x30);
	write_reg(0x34, 0x9b);
	write_reg(0x36, 0xc0);
	write_reg(0xc8, 0x92);
	write_reg(0xca, 0xd3);
	write_reg(0xcc, 0xd3);
	write_reg(0xd8, 0x60);
	write_reg(0x38, 0x00);
	write_reg(0x76, 0x10);
	write_reg(0x78, 0x80);
	write_reg(0x7a, 0x72);
	write_reg(0x7c, 0x82);
	write_reg(0x7e, 0x42);
	write_reg(0x80, 0xF5);
	write_reg(0x82, 0x13);
	write_reg(0x84, 0xF2);
	write_reg(0x86, 0x26);
	write_reg(0x88, 0x00);
	write_reg(0x8a, 0x08);
	write_reg(0x8c, 0x77);
	write_reg(0x8e, 0x03);
	write_reg(0x90, 0x0D);
	write_reg(0x92, 0x24);
	write_reg(0x94, 0xE0);
	write_reg(0x96, 0x06);
	write_reg(0x98, 0x00);
	write_reg(0x9a, 0x10);
	write_reg(0x9c, 0x68);
	write_reg(0x9e, 0xDA);
	write_reg(0xa0, 0x0A);
	write_reg(0xa2, 0x0A);
	write_reg(0xa4, 0xE5);
	write_reg(0xa6, 0x77);
	write_reg(0xa8, 0xC2);
	write_reg(0xaa, 0x8A);
	write_reg(0xac, 0x9A);
	write_reg(0xae, 0x12);
	write_reg(0xb0, 0x99);
	write_reg(0xb2, 0x86);
	write_reg(0xb4, 0x25);
	write_reg(0xb6, 0x00);
	write_reg(0xbc, 0x00);
	write_reg(0xbe, 0x00);
	write_reg(0xc0, 0x00);
	write_reg(0xc2, 0x00);
	mdelay(1);
	write_reg(0x6c, 0xc4);
	write_reg(0xc6, 0x03);
	write_reg(0xc4, 0x01);
	write_reg(0x32, 0x00);
	register_reboot_notifier(&conexant_cx25871_notifier);
}

static void
conexant_cx25871_on(struct fb_info *info)
{
	write_reg(0xba, 0x20);
}

static void
conexant_cx25871_off(struct fb_info *info)
{
	write_reg(0xba, 0x30);
}

typedef enum
{
	CRT_VGA_350, //CRT_VGA_350_70
	CRT_AIC_350_50, //CRT_VGA_350_50
	CRT_AIC_350_60, //CRT_VGA_350_60
	CRT_AIC_400_50, //CRT_VGA_400_50
	CRT_AIC_400_60, //CRT_VGA_400_60
	CRT_PC9801,
	CRT_VGA_GRAPH, //CRT_VGA_400_GRAPH, CRT_VGA_400, CRT_VGA_400_70
	CRT_PC9821,
	CRT_VESA_PAL, //CRT_VGA_480_50
	CRT_VGA_480, //CRT_VGA_480_60, CRT_VGA, CRT_VGA_60
	CRT_MAC_480,
	CRT_VESA_480_72, //CRT_VGA_480_72, CRT_VGA_72
	CRT_VESA_480_75, //CRT_VGA_480_75, CRT_VGA_75
	CRT_VGA_400_TEXT, //CRT_VGA_TEXT
	CRT_AIC_720_400_50,
	CRT_AIC_720_400_60,
	CRT_MAC_800,
	CRT_SVGA, //CRT_SVGA_56
	CRT_VESA_600_60, //CRT_SVGA_60
	CRT_VESA_600_72, //CRT_SVGA_72
	CRT_VESA_600_75, //CRT_SVGA_75
	CRT_XGA, //CRT_XGA_60
	CRT_COMPAQ_XGA,
	CRT_VESA_768_70, //CRT_XGA_70
	CRT_VESA_768_75, //CRT_XGA_75
	CRT_MAC_768,
	CRT_XGAPLUS, //CRT_XGAPLUS_60
	CRT_XGAPLUS_70,
	CRT_XGAPLUS_75,
	CRT_VG828_900_66,
	CRT_VG828_900_76,
	CRT_QUADVGA, //CRT_QUADVGA_60
	CRT_VESA_1024_60, //CRT_SXGA, CRT_SXGA_60
	PHILIPS_LB064V02,
	CX25871_NTSC,
	CX25871_OVERSCAN,
	CX25871_PAL,
} DisplayType;

static struct ep93xxfb_videotimings modedb[] = {
	[CRT_VGA_350] = {
		.name[0] =	"CRT-VGA-350",
		.name[1] =	"CRT-VGA-350@70",
		.name[2] =	"CRT-640x350",
		.name[3] =	"CRT-640x350@70",
		.name[4] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		350,
		.vfp =		37,
		.vsync =	2,
		.vbp =		60,
		.refresh =	70, //70.087
		.mrefresh =	87,
		.width =	640,
		.height =	350,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_LOW,
	},
	[CRT_AIC_350_50] = {
		.name[0] =	"CRT-AIC-350@50",
		.name[1] =	"CRT-VGA-350@50",
		.name[2] =	"CRT-640x350@50",
		.name[3] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		350,
		.vfp =		127,
		.vsync =	2,
		.vbp =		150,
		.refresh =	50, //50.030
		.mrefresh =	30,
		.width =	640,
		.height =	350,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_LOW,
	},
	[CRT_AIC_350_60] = {
		.name[0] =	"CRT-AIC-350@60",
		.name[1] =	"CRT-VGA-350@60",
		.name[2] =	"CRT-640x350@60",
		.name[3] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		350,
		.vfp =		75,
		.vsync =	2,
		.vbp =		98,
		.refresh =	59, //59.940
		.mrefresh =	940,
		.width =	640,
		.height =	350,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_LOW,
	},
	[CRT_AIC_400_50] = {
		.name[0] =	"CRT-AIC-400@50",
		.name[1] =	"CRT-VGA-400@50",
		.name[2] =	"CRT-640x400@50",
		.name[3] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		400,
		.vfp =		102,
		.vsync =	2,
		.vbp =		125,
		.refresh =	50, //50.030
		.mrefresh =	30,
		.width =	640,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_AIC_400_60] = {
		.name[0] =	"CRT-AIC-400@60",
		.name[1] =	"CRT-VGA-400@60",
		.name[2] =	"CRT-640x400@60",
		.name[3] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		400,
		.vfp =		50,
		.vsync =	2,
		.vbp =		73,
		.refresh =	59, //59.940
		.mrefresh =	940,
		.width =	640,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_PC9801] = {
		.name[0] =	"CRT-PC9801",
		.name[1] =	NULL,
		.hres =		640,
		.hfp =		64,
		.hsync =	64,
		.hbp =		80,
		.vres =		400,
		.vfp =		7,
		.vsync =	8,
		.vbp =		25,
		.refresh =	56, //56.420
		.mrefresh =	420,
		.width =	640,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VGA_GRAPH] = {
		.name[0] =	"CRT-VGA-GRAPH",
		.name[1] =	"CRT-VGA-400-GRAPH",
		.name[2] =	"CRT-VGA-400",
		.name[3] =	"CRT-VGA-400@70",
		.name[4] =	"CRT-640x400",
		.name[5] =	"CRT-640x400@70",
		.name[6] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		400,
		.vfp =		12,
		.vsync =	2,
		.vbp =		35,
		.refresh =	70, //70.087
		.mrefresh =	87,
		.width =	640,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_PC9821] = {
		.name[0] =	"CRT-PC9821",
		.name[1] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	64,
		.hbp =		80,
		.vres =		400,
		.vfp =		13,
		.vsync =	2,
		.vbp =		34,
		.refresh =	70, //70.150
		.mrefresh =	150,
		.width =	640,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VESA_PAL] = {
		.name[0] =	"CRT-VESA-PAL",
		.name[1] =	"CRT-VGA-480@50",
		.name[2] =	"CRT-VGA@50",
		.name[3] =	"CRT-640x480@50",
		.name[4] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		480,
		.vfp =		62,
		.vsync =	2,
		.vbp =		85,
		.refresh =	50, //50.030
		.mrefresh =	30,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VGA_480] = {
		.name[0] =	"CRT-VGA-480",
		.name[1] =	"CRT-VGA-480@60",
		.name[2] =	"CRT-VGA",
		.name[3] =	"CRT-VGA@60",
		.name[4] =	"CRT-640x480",
		.name[5] =	"CRT-640x480@60",
		.name[6] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		480,
		.vfp =		10,
		.vsync =	2,
		.vbp =		33,
		.refresh =	59, //59.940
		.mrefresh =	940,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_MAC_480] = {
		.name[0] =	"CRT-MAC-480",
		.name[1] =	NULL,
		.hres =		640,
		.hfp =		64,
		.hsync =	64,
		.hbp =		96,
		.vres =		480,
		.vfp =		3,
		.vsync =	3,
		.vbp =		39,
		.refresh =	66, //66.670
		.mrefresh =	670,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VESA_480_72] = {
		.name[0] =	"CRT-VESA-480@72",
		.name[1] =	"CRT-VGA-480@72",
		.name[2] =	"CRT-VGA@72",
		.name[3] =	"CRT-640x480@72",
		.name[4] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	40,
		.hbp =		120,
		.vres =		480,
		.vfp =		1,
		.vsync =	3,
		.vbp =		20,
		.refresh =	72, //72.809
		.mrefresh =	809,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VESA_480_75] = {
		.name[0] =	"CRT-VESA-480@75",
		.name[1] =	"CRT-VGA-480@75",
		.name[2] =	"CRT-VGA@75",
		.name[3] =	"CRT-640x480@75",
		.name[4] =	NULL,
		.hres =		640,
		.hfp =		16,
		.hsync =	64,
		.hbp =		120,
		.vres =		480,
		.vfp =		1,
		.vsync =	3,
		.vbp =		16,
		.refresh =	75, //75.000
		.mrefresh =	0,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VGA_400_TEXT] = {
		.name[0] =	"CRT-VGA-400-TEXT",
		.name[1] =	"CRT-VGA-TEXT",
		.name[2] =	"CRT-720x400",
		.name[3] =	"CRT-720x400@70",
		.name[4] =	NULL,
		.hres =		720,
		.hfp =		18,
		.hsync =	108,
		.hbp =		54,
		.vres =		400,
		.vfp =		12,
		.vsync =	2,
		.vbp =		35,
		.refresh =	70, //70.087
		.mrefresh =	87,
		.width =	720,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_AIC_720_400_50] = {
		.name[0] =	"CRT-AIC-720x400@50",
		.name[1] =	"CRT-720x400@50",
		.name[2] =	NULL,
		.hres =		720,
		.hfp =		18,
		.hsync =	108,
		.hbp =		54,
		.vres =		400,
		.vfp =		102,
		.vsync =	2,
		.vbp =		125,
		.refresh =	50, //50.030
		.mrefresh =	30,
		.width =	720,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_AIC_720_400_60] = {
		.name[0] =	"CRT-AIC-720x400@60",
		.name[1] =	"CRT-720x400@60",
		.name[2] =	NULL,
		.hres =		720,
		.hfp =		18,
		.hsync =	108,
		.hbp =		54,
		.vres =		400,
		.vfp =		50,
		.vsync =	2,
		.vbp =		73,
		.refresh =	59, //59.940
		.mrefresh =	940,
		.width =	720,
		.height =	400,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_MAC_800] = {
		.name[0] =	"CRT-MAC-800",
		.name[1] =	NULL,
		.hres =		832,
		.hfp =		32,
		.hsync =	64,
		.hbp =		224,
		.vres =		624,
		.vfp =		1,
		.vsync =	3,
		.vbp =		39,
		.refresh =	74, //74.550
		.mrefresh =	550,
		.width =	832,
		.height =	624,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_SVGA] = {
		.name[0] =	"CRT-SVGA",
		.name[1] =	"CRT-SVGA@56",
		.name[2] =	"CRT-800x600",
		.name[3] =	"CRT-800x600@56",
		.name[4] =	NULL,
		.hres =		800,
		.hfp =		24,
		.hsync =	72,
		.hbp =		128,
		.vres =		600,
		.vfp =		1,
		.vsync =	2,
		.vbp =		22,
		.refresh =	56, //56.250
		.mrefresh =	250,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VESA_600_60] = {
		.name[0] =	"CRT-VESA-600@60",
		.name[1] =	"CRT-SVGA@60",
		.name[2] =	"CRT-800x600@60",
		.name[3] =	NULL,
		.hres =		800,
		.hfp =		40,
		.hsync =	128,
		.hbp =		88,
		.vres =		600,
		.vfp =		1,
		.vsync =	4,
		.vbp =		23,
		.refresh =	60, //60.317
		.mrefresh =	317,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VESA_600_72] = {
		.name[0] =	"CRT-VESA-600@72",
		.name[1] =	"CRT-SVGA@72",
		.name[2] =	"CRT-800x600@72",
		.name[3] =	NULL,
		.hres =		800,
		.hfp =		56,
		.hsync =	120,
		.hbp =		64,
		.vres =		600,
		.vfp =		37,
		.vsync =	6,
		.vbp =		23,
		.refresh =	72, //72.188
		.mrefresh =	188,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VESA_600_75] = {
		.name[0] =	"CRT-VESA-600@75",
		.name[1] =	"CRT-SVGA@75",
		.name[2] =	"CRT-800x600@75",
		.name[3] =	NULL,
		.hres =		800,
		.hfp =		16,
		.hsync =	80,
		.hbp =		160,
		.vres =		600,
		.vfp =		1,
		.vsync =	3,
		.vbp =		21,
		.refresh =	75, //75.000
		.mrefresh =	0,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGA] = {
		.name[0] =	"CRT-XGA",
		.name[1] =	"CRT-XGA@60",
		.name[2] =	"CRT-1024x768",
		.name[3] =	"CRT-1024x768@60",
		.name[4] =	NULL,
		.hres =		1024,
		.hfp =		24,
		.hsync =	136,
		.hbp =		160,
		.vres =		768,
		.vfp =		3,
		.vsync =	6,
		.vbp =		29,
		.refresh =	60, //60.004
		.mrefresh =	4,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_COMPAQ_XGA] = {
		.name[0] =	"CRT-COMPAQ-XGA",
		.name[1] =	NULL,
		.hres =		1024,
		.hfp =		16,
		.hsync =	176,
		.hbp =		112,
		.vres =		768,
		.vfp =		8,
		.vsync =	4,
		.vbp =		36,
		.refresh =	66, //66.132
		.mrefresh =	132,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VESA_768_70] = {
		.name[0] =	"CRT-VESA-768@70",
		.name[1] =	"CRT-XGA@70",
		.name[2] =	"CRT-1024x768@70",
		.name[3] =	NULL,
		.hres =		1024,
		.hfp =		24,
		.hsync =	136,
		.hbp =		144,
		.vres =		768,
		.vfp =		3,
		.vsync =	6,
		.vbp =		29,
		.refresh =	70, //70.069
		.mrefresh =	69,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VESA_768_75] = {
		.name[0] =	"CRT-VESA-768@75",
		.name[1] =	"CRT-XGA@75",
		.name[2] =	"CRT-1024x768@75",
		.name[3] =	NULL,
		.hres =		1024,
		.hfp =		16,
		.hsync =	96,
		.hbp =		176,
		.vres =		768,
		.vfp =		1,
		.vsync =	3,
		.vbp =		28,
		.refresh =	75, //75.029
		.mrefresh =	29,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_MAC_768] = {
		.name[0] =	"CRT-MAC-768",
		.name[1] =	NULL,
		.hres =		1024,
		.hfp =		32,
		.hsync =	96,
		.hbp =		176,
		.vres =		768,
		.vfp =		3,
		.vsync =	3,
		.vbp =		29,
		.refresh =	75, //75.020
		.mrefresh =	20,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_XGAPLUS] = {
		.name[0] =	"CRT-XGA+",
		.name[1] =	"CRT-XGA+@60",
		.name[2] =	"CRT-1152x864",
		.name[3] =	"CRT-1152x864@60",
		.name[4] =	NULL,
		.hres =		1152,
		.hfp =		40,
		.hsync =	96,
		.hbp =		192,
		.vres =		864,
		.vfp =		13,
		.vsync =	3,
		.vbp =		32,
		.refresh =	59, //59.270
		.mrefresh =	270,
		.width =	1152,
		.height =	864,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGAPLUS_70] = {
		.name[0] =	"CRT-XGA+@70",
		.name[1] =	"CRT-1152x864@70",
		.name[2] =	NULL,
		.hres =		1152,
		.hfp =		32,
		.hsync =	96,
		.hbp =		200,
		.vres =		864,
		.vfp =		1,
		.vsync =	3,
		.vbp =		44,
		.refresh =	70, //70.012
		.mrefresh =	12,
		.width =	1152,
		.height =	864,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGAPLUS_75] = {
		.name[0] =	"CRT-XGA+@75",
		.name[1] =	"CRT-1152x864@75",
		.name[2] =	NULL,
		.hres =		1152,
		.hfp =		64,
		.hsync =	128,
		.hbp =		138, //256,
		.vres =		864,
		.vfp =		2,
		.vsync =	2,
		.vbp =		32,
		.refresh =	75, //75.000
		.mrefresh =	0,
		.width =	1152,
		.height =	864,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VG828_900_66] = {
		.name[0] =	"CRT-VG828-900@66",
		.name[1] =	NULL,
		.hres =		1152,
		.hfp =		30,
		.hsync =	128,
		.hbp =		194,
		.vres =		900,
		.vfp =		2,
		.vsync =	4,
		.vbp =		31,
		.refresh =	65, //65.950
		.mrefresh =	950,
		.width =	1152,
		.height =	900,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VG828_900_76] = {
		.name[0] =	"CRT-VG828-900@76",
		.name[1] =	NULL,
		.hres =		1152,
		.hfp =		16,
		.hsync =	96,
		.hbp =		130, //208
		.vres =		900,
		.vfp =		2,
		.vsync =	8,
		.vbp =		33,
		.refresh =	76, //76.068
		.mrefresh =	68,
		.width =	1152,
		.height =	900,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_QUADVGA] = {
		.name[0] =	"CRT-QuadVGA",
		.name[1] =	"CRT-QuadVGA@60",
		.name[2] =	"CRT-1280x960",
		.name[3] =	"CRT-1280x960@60",
		.name[4] =	NULL,
		.hres =		1280,
		.hfp =		96,
		.hsync =	112,
		.hbp =		178, //312
		.vres =		960,
		.vfp =		1,
		.vsync =	3,
		.vbp =		36,
		.refresh =	60, //60.000
		.mrefresh =	0,
		.width =	1280,
		.height =	960,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_VESA_1024_60] = {
		.name[0] =	"CRT-VESA-1024@60",
		.name[1] =	"CRT-SXGA",
		.name[2] =	"CRT-SXGA@60",
		.name[3] =	"CRT-1280x1024",
		.name[4] =	"CRT-1280x1024@60",
		.name[5] =	NULL,
		.hres =		1280,
		.hfp =		48,
		.hsync =	112,
		.hbp =		124, //248
		.vres =		1024,
		.vfp =		1,
		.vsync =	3,
		.vbp =		38,
		.refresh =	60, //60.000
		.mrefresh =	0,
		.width =	1280,
		.height =	1024,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[PHILIPS_LB064V02] = {
		.name[0] =	"Philips-LB064V02",
		.name[1] =	NULL,
		.on =		philips_lb064v02_on,
		.off =		philips_lb064v02_off,
		.hres =		640,
		.hfp =		24,
		.hsync =	96,
		.hbp =		40,
		.vres =		480,
		.vfp =		10,
		.vsync =	2,
		.vbp =		33,
		.refresh =	60,
		.mrefresh =	0,
		.width =	132,
		.height =	100,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_NTSC] = {
		.name[0] = 	"Conexant-CX25871-NTSC",
		.name[1] =	NULL,
		.configure =	conexant_cx25871_configure_ntsc,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		16,
		.hsync =	4,
		.hbp =		124,
		.vres =		480,
		.vfp =		45,
		.vsync =	2,
		.vbp =		73,
		.refresh =	35466,
		.mrefresh =	0,
		.width =	300,
		.height =	230,
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_OVERSCAN] = {
		.name[0] =	"Conexant-CX25871-overscan",
		.name[1] =	NULL,
		.configure =	conexant_cx25871_overscan_configure,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		15,
		.hsync =	4,
		.hbp =		117,
		.vres =		480,
		.vfp =		9,
		.vsync =	2,
		.vbp =		34,
		.refresh =	40950,
		.mrefresh =	0,
		.width =	300,
		.height =	230,
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_PAL] = {
		.name[0] =	"Conexant-CX25871-PAL",
		.name[1] =	NULL,
		.configure =	conexant_cx25871_configure_pal,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		50, //50
		.hsync =	4,//4 
		.hbp =		250, //250
		.vres =		480,
		.vfp =		50,
		.vsync =	2,
		.vbp =		93,
		.refresh =	35466,
		.mrefresh =	0,
		.width =	320, //300
		.height =	240, //230
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING, 
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
};

#if defined(CONFIG_FB_EP93XX_CRT)
#if defined (CONFIG_FB_EP93XX_CRT_VGA_350_50)
#define DEFAULT_MODE CRT_AIC_350_50
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_350_60)
#define DEFAULT_MODE CRT_AIC_350_60
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_350_70)
#define DEFAULT_MODE CRT_VGA_350
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_400_50)
#define DEFAULT_MODE CRT_AIC_400_50
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_400_60)
#define DEFAULT_MODE CRT_AIC_400_60
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_400_70)
#define DEFAULT_MODE CRT_VGA_GRAPH
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_50)
#define DEFAULT_MODE CRT_VESA_PAL
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_60)
#define DEFAULT_MODE CRT_VGA_480
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_72)
#define DEFAULT_MODE CRT_VESA_480_72
#elif defined (CONFIG_FB_EP93XX_CRT_VGA_75)
#define DEFAULT_MODE CRT_VESA_480_75
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA_56)
#define DEFAULT_MODE CRT_SVGA
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA_60)
#define DEFAULT_MODE CRT_VESA_600_60
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA_72)
#define DEFAULT_MODE CRT_VESA_600_72
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA_75)
#define DEFAULT_MODE CRT_VESA_600_75
#elif defined (CONFIG_FB_EP93XX_CRT_XGA_60)
#define DEFAULT_MODE CRT_XGA
#elif defined (CONFIG_FB_EP93XX_CRT_XGA_70)
#define DEFAULT_MODE CRT_VESA_768_70
#elif defined (CONFIG_FB_EP93XX_CRT_XGA_75)
#define DEFAULT_MODE CRT_VESA_768_75
#elif defined (CONFIG_FB_EP93XX_CRT_XGAPLUS_60)
#define DEFAULT_MODE CRT_XGAPLUS
#elif defined (CONFIG_FB_EP93XX_CRT_XGAPLUS_70)
#define DEFAULT_MODE CRT_XGAPLUS_70
#elif defined (CONFIG_FB_EP93XX_CRT_XGAPLUS_75)
#define DEFAULT_MODE CRT_XGAPLUS_75
#elif defined (CONFIG_FB_EP93XX_CRT_QUADVGA_60)
#define DEFAULT_MODE CRT_QUADVGA
#elif defined (CONFIG_FB_EP93XX_CRT_SXGA_60)
#define DEFAULT_MODE CRT_VESA_1024_60
#else
#error What Display Setting was that!!!
#endif
#elif defined(CONFIG_FB_EP93XX_LCD)
#define DEFAULT_MODE PHILIPS_LB064V02
#elif defined(CONFIG_FB_EP93XX_PAL)
#define DEFAULT_MODE CX25871_PAL
#elif defined(CONFIG_FB_EP93XX_NTSC)
#ifdef CONFIG_FB_EP93XX_NTSC_OVERSCAN
#define DEFAULT_MODE CX25871_OVERSCAN
#else
#define DEFAULT_MODE CX25871_NTSC
#endif
#else
#define DEFAULT_MODE CRT_VGA
#endif
#if defined(CONFIG_FB_EP93XX_8BPP)
#define DEFAULT_BPP  8
#define DEFAULT_BYTES_PER_PIXEL 1
#define DEFAULT_PIXEL_MASK 3
#define DEFAULT_PIXEL_SHIFT 2
#define DEFAULT_PIXEL_FORMAT 0x00040000
#elif defined(CONFIG_FB_EP93XX_16BPP)
#define DEFAULT_BPP 16
#define DEFAULT_BYTES_PER_PIXEL 2
#define DEFAULT_PIXEL_MASK 1
#define DEFAULT_PIXEL_SHIFT 1
#define DEFAULT_PIXEL_FORMAT 0x00080000
#elif defined(CONFIG_FB_EP93XX_24BPP)
#define DEFAULT_BPP 24
#define DEFAULT_BYTES_PER_PIXEL 3
#define DEFAULT_PIXEL_MASK 0
#define DEFAULT_PIXEL_SHIFT 0
#define DEFAULT_PIXEL_FORMAT 0x00100000
#elif defined(CONFIG_FB_EP93XX_32BPP)
#define DEFAULT_BPP 32
#define DEFAULT_BYTES_PER_PIXEL 3
#define DEFAULT_PIXEL_MASK 0
#define DEFAULT_PIXEL_SHIFT 0
#define DEFAULT_PIXEL_FORMAT 0x00100000
#else
#define DEFAULT_BPP 16
#define DEFAULT_BYTES_PER_PIXEL 2
#define DEFAULT_PIXEL_MASK 1
#define DEFAULT_PIXEL_SHIFT 1
#define DEFAULT_PIXEL_FORMAT 0x00080000
#endif

static unsigned int pseudo_palette[256];

#ifdef MODULE
static char *modestr = 0;
#endif
static int mode = DEFAULT_MODE;
static int bpp = DEFAULT_BPP;
static int bytes_per_pixel = DEFAULT_BYTES_PER_PIXEL;
static int pixel_mask = DEFAULT_PIXEL_MASK;
static int pixel_shift = DEFAULT_PIXEL_SHIFT;
static int pixel_format = DEFAULT_PIXEL_FORMAT;

#define ep93xxfb_outl(value, reg)				\
	{							\
		outl(RASTER_SWLOCK_VALUE, RASTER_SWLOCK);	\
		outl(value, reg);				\
	}

#ifdef CONFIG_EP93XX_GRAPHICS

static DECLARE_WAIT_QUEUE_HEAD(ep93xxfb_wait_in);

static irqreturn_t
ep93xxfb_irq_handler(int i, void *blah, struct pt_regs *regs)
{
	outl(0x00000000, BLOCKCTRL);

        wake_up_interruptible(&ep93xxfb_wait_in);
        
        return IRQ_HANDLED; 
}

static void
ep93xxfb_wait(void)
{
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(&ep93xxfb_wait_in, &wait);
	current->state = TASK_INTERRUPTIBLE;
	while (inl(BLOCKCTRL) & 0x00000001) {
		schedule();
		current->state = TASK_INTERRUPTIBLE;
	}
	remove_wait_queue(&ep93xxfb_wait_in, &wait);
	current->state = TASK_RUNNING;
}

static int
ep93xxfb_line(struct fb_info *info,  struct ep93xx_line *line)
{
	unsigned long value = 0;
        long dx, dy, count, xinc, yinc, xval, yval, incr;

        if ((line->x1 >= info->var.xres) ||
            (line->x2 >= info->var.xres) ||
            (line->y1 >= info->var.yres) ||
            (line->y2 >= info->var.yres))
		return -EFAULT;

	dx = line->x2 - line->x1;
	if (dx < 0) {
		value |= 0x00000020;
		dx *= -1;
	}
	dy = line->y2 - line->y1;
	if (dy < 0) {
		value |= 0x00000040;
		dy *= -1;
	}
	if (line->flags & LINE_PRECISE) {
		count = isqrt(((dy * dy) + (dx * dx)) * 4096);
		xinc = (4095 * 64 * dx) / count;
		yinc = (4095 * 64 * dy) / count;
		xval = 2048;
		yval = 2048;
		count = 0;
		while (dx || dy) {
			incr = 0;
			xval -= xinc;
			if (xval <= 0) {
				xval += 4096;
				dx--;
				incr = 1;
			}
			yval -= yinc;
			if (yval <= 0) {
				yval += 4096;
				dy--;
				incr = 1;
			}
			count += incr;
		}
	} else {
		if (dx == dy) {
			xinc = 4095;
			yinc = 4095;
			count = dx;
		} else if (dx < dy) {
			xinc = (dx * 4095) / dy;
			yinc = 4095;
			count = dy;
		} else {
			xinc = 4095;
			yinc = (dy * 4095) / dx;
			count = dx;
		}
	}

	outl(0x08000800, LINEINIT);
	if (line->flags & LINE_PATTERN)
		outl(line->pattern, LINEPATTRN);
	else
		outl(0x000fffff, LINEPATTRN);
	outl(fb_phys + (line->y1 *(info->var.xres * bytes_per_pixel)) +
	     ((line->x1 * bytes_per_pixel) & ~pixel_mask), BLKDSTSTRT);
	outl(((line->x1 & pixel_mask) * bpp) |
	     ((((line->x1 + dx - 1) & pixel_mask) *
	       bpp) << 16), DESTPIXELSTRT);
	outl((info->var.xres * bytes_per_pixel)/ 4, DESTLINELENGTH);
	outl(line->fgcolor, BLOCKMASK);
	outl(line->bgcolor, BACKGROUND);
	outl((yinc << 16) | xinc, LINEINC);
	outl(count & 0xfff, BLKDESTWIDTH);
	outl(0, BLKDESTHEIGHT);
	value |= (line->flags & LINE_BACKGROUND) ? 0x00004000 : 0;
	outl(value | pixel_format | 0x00000013, BLOCKCTRL);
	ep93xxfb_wait();
	return 0;
}

static int
ep93xxfb_fill(struct fb_info *info, struct ep93xx_fill *fill)
{
	if ((fill->dx >= info->var.xres) ||
	    (fill->dy >= info->var.yres))
		return -EFAULT;

	if ((fill->dx + fill->width - 1) >= info->var.xres)
		fill->width = info->var.xres - fill->dx;
	if ((fill->dy + fill->height - 1) >= info->var.yres)
		fill->height = info->var.yres - fill->dy;

	outl(fb_phys + (fill->dy *(info->var.xres * bytes_per_pixel)) +
	     ((fill->dx * bytes_per_pixel) & ~pixel_mask), BLKDSTSTRT);
	outl(((fill->dx & pixel_mask) * bpp) |
	     ((((fill->dx + fill->width - 1) & pixel_mask) *
	       bpp) << 16), DESTPIXELSTRT);
	outl((info->var.xres * bytes_per_pixel)/ 4, DESTLINELENGTH);
	outl(((fill->dx + fill->width - 1) >> pixel_shift) -
	     (fill->dx >> pixel_shift), BLKDESTWIDTH);
	outl(fill->height - 1, BLKDESTHEIGHT);
	outl(fill->color, BLOCKMASK);
	outl(pixel_format | 0x0000000b, BLOCKCTRL);

	ep93xxfb_wait();
	return 0;
}


static unsigned char pucBlitBuf[4096];

static int
ep93xxfb_blit(struct fb_info *info, struct ep93xx_blit *blit)
{
	unsigned long value = 0, size, dx1, dx2, dy1, sx1, sx2, sy1;

	if ((blit->dx >= info->var.xres) ||
	    (blit->dy >= info->var.yres) ||
	    ((blit->dx + blit->width - 1) >= info->var.xres) ||
	    ((blit->dy + blit->height - 1) >= info->var.yres))
		return -EFAULT;

	value = blit->flags & (BLIT_TRANSPARENT | BLIT_MASK_MASK |
			       BLIT_DEST_MASK | BLIT_1BPP_SOURCE);

	if ((blit->flags & BLIT_SOURCE_MASK) == BLIT_SOURCE_MEMORY) {
		if (blit->flags & BLIT_1BPP_SOURCE) {
			size = blit->swidth * blit->height / 8;
			if (size <= 4096)
				copy_from_user(pucBlitBuf, blit->data, size);
			else {
				for (size = blit->height; size; ) {
					blit->height = 32768 / blit->swidth;
					if (blit->height > size)
						blit->height = size;
					ep93xxfb_blit(info, blit);
					blit->dy += blit->height;
					blit->data += blit->swidth *
						      blit->height / 8;
					size -= blit->height;
				}
				return 0;
			}
			if (blit->flags & BLIT_TRANSPARENT)
				value ^= 0x00004000;
			outl(0x00000007, SRCPIXELSTRT);
			outl(blit->swidth >> 5, SRCLINELENGTH);
			outl((blit->width - 1) >> 5, BLKSRCWIDTH);
			outl(blit->bgcolor, BACKGROUND);
		} else {
			size = blit->swidth * blit->height * bytes_per_pixel;
			if (size <= 4096)
				copy_from_user(pucBlitBuf, blit->data, size);
			else {
				for (size = blit->height; size; ) {
					blit->height = 4096 / bytes_per_pixel /
						blit->swidth;
					if (blit->height > size)
						blit->height = size;
					ep93xxfb_blit(info,blit);
					blit->dy += blit->height;
					blit->data += blit->swidth *
						      blit->height *
						      bytes_per_pixel;
					size -= blit->height;
				}
				return 0;
			}
			outl(0x00000000, SRCPIXELSTRT);
			outl(blit->swidth >> pixel_shift, SRCLINELENGTH);
			outl((blit->width - 1) >> pixel_shift, BLKSRCWIDTH);
		}
		outl(virt_to_phys(pucBlitBuf), BLKSRCSTRT);
		dx1 = blit->dx;
		dy1 = blit->dy;
		dx2 = blit->dx + blit->width - 1;
	} else {					//SOURCE_SCREEN
		if ((blit->sx >= info->var.xres) ||
		    (blit->sy >= info->var.yres) ||
		    ((blit->sx + blit->width - 1) >= info->var.xres) ||
		    ((blit->sy + blit->height - 1) >= info->var.yres))
			return -EFAULT;

		if ((blit->dy == blit->sy) && (blit->dx == blit->sx))
			return 0;

		if ((blit->dy == blit->sy) &&
		    (blit->dx > blit->sx) &&
		    (blit->dx < (blit->sx + blit->width - 1))) {
			dx1 = blit->dx + blit->width - 1;
			dx2 = blit->dx;
			sx1 = blit->sx + blit->width - 1;
			sx2 = blit->sx;
			value |= 0x000000a0;
		} else {
			dx1 = blit->dx;
			dx2 = blit->dx + blit->width - 1;
			sx1 = blit->sx;
			sx2 = blit->sx + blit->width - 1;
		}

		if (blit->dy <= blit->sy) {
			dy1 = blit->dy;
			sy1 = blit->sy;
		} else {
			dy1 = blit->dy + blit->height - 1;
			sy1 = blit->sy + blit->height - 1;
			value |= 0x00000140;
		}

		outl(((sx1 & pixel_mask) * bpp) |
		     (((sx2 & pixel_mask) * bpp) << 16),
		     SRCPIXELSTRT);

		outl(fb_phys + (sy1 *(info->var.xres * bytes_per_pixel)) +
		     ((sx1 * bytes_per_pixel) & ~pixel_mask), BLKSRCSTRT);

		outl((info->var.xres * bytes_per_pixel)/ 4, SRCLINELENGTH);
		if (sx1 < sx2)
			outl((sx2 >> pixel_shift) - (sx1 >> pixel_shift),
			     BLKSRCWIDTH);
		else
			outl((sx1 >> pixel_shift) - (sx2 >> pixel_shift),
			     BLKSRCWIDTH);
	}

	outl(((dx1 & pixel_mask) * bpp) |
	     (((dx2 & pixel_mask) * bpp) << 16), DESTPIXELSTRT);
	outl(fb_phys + (dy1 *(info->var.xres * bytes_per_pixel)) +
             ((dx1 * bytes_per_pixel) & ~pixel_mask), BLKDSTSTRT);

	outl((info->var.xres * bytes_per_pixel)/ 4, DESTLINELENGTH);
	if (dx1 < dx2)
		outl((dx2 >> pixel_shift) - (dx1 >> pixel_shift),
		     BLKDESTWIDTH);
	else
		outl((dx1 >> pixel_shift) - (dx2 >> pixel_shift),
		     BLKDESTWIDTH);
	outl(blit->height - 1, BLKDESTHEIGHT);
	outl(blit->fgcolor, BLOCKMASK);
	outl(blit->transcolor, TRANSPATTRN);
	outl(value | pixel_format | 0x00000003, BLOCKCTRL);
	ep93xxfb_wait();
	return 0;
}

#endif

static unsigned long cursor_data[64][4];

static int
ep93xxfb_cursor(struct fb_info *info, struct ep93xx_cursor *cursor)
{
	unsigned long data[64 * 4];
	long i, x, y, save;

	if (cursor->flags & CURSOR_OFF)
		outl(inl(CURSORXYLOC) & ~0x00008000, CURSORXYLOC);

	if (cursor->flags & CURSOR_SETSHAPE) {
		copy_from_user(data, cursor->data,
			       cursor->width * cursor->height / 4);

		save = inl(CURSORXYLOC);
		outl(save & ~0x00008000, CURSORXYLOC);

		for (y = 0, i = 0; y < cursor->height; y++) {
			for (x = 0; x < cursor->width; x += 16)
				cursor_data[y][x] = data[i++];
		}

		outl(virt_to_phys(cursor_data), CURSOR_ADR_START);
		outl(virt_to_phys(cursor_data), CURSOR_ADR_RESET);
		outl(0x00000300 | ((cursor->height - 1) << 2) |
		     ((cursor->width - 1) >> 4), CURSORSIZE);
		outl(save, CURSORXYLOC);
	}

	if (cursor->flags & CURSOR_SETCOLOR) {
		outl(cursor->color1, CURSORCOLOR1);
		outl(cursor->color2, CURSORCOLOR2);
		outl(cursor->blinkcolor1, CURSORBLINK1);
		outl(cursor->blinkcolor2, CURSORBLINK2);
	}

	if (cursor->flags & CURSOR_BLINK) {
		if (cursor->blinkrate)
			outl(0x00000100 | cursor->blinkrate, CURSORBLINK);
		else
			outl(0x000000ff, CURSORBLINK);
	}

	if (cursor->flags & CURSOR_MOVE) {
		x = (inl(HACTIVESTRTSTOP) & 0x000003ff) - cursor->dx - 2;
		y = (inl(VACTIVESTRTSTOP) & 0x000003ff) - cursor->dy;
		outl((inl(CURSORXYLOC) & 0x8000) | (y << 16) | x, CURSORXYLOC);
	}

	if (cursor->flags & CURSOR_ON)
		outl(inl(CURSORXYLOC) | 0x00008000, CURSORXYLOC);
		
	return 0;
}

static inline void
ep93xxfb_palette_write(u_int regno, u_int red, u_int green,
		       u_int blue, u_int trans)
{
    unsigned int cont, i, pal;

    pal = ((red & 0xFF00) << 8) | (green & 0xFF00) | ((blue & 0xFF00) >> 8);

    pseudo_palette[regno] = pal;

    outl( pal, (COLOR_LUT+(regno<<2)) );

    cont = inl(LUTCONT);

    if ((cont&LUTCONT_STAT && cont&LUTCONT_RAM1) ||
	(!(cont&LUTCONT_STAT) && !(cont&LUTCONT_RAM1)))
    {
       	for (i=0; i< 256; i++)  // Update inactive LUT
	{
	    outl( pseudo_palette[i], (COLOR_LUT+(i<<2)) );
	}
	// Switch active LUTs next frame
	outl( cont ^ LUTCONT_RAM1, LUTCONT );
    }
}

static int
ep93xxfb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue,
		   unsigned transp, struct fb_info *info)
{

#define CNVT_TOHW(val,width) ((((val)<<(width))+0x7FFF-(val))>>16)

	switch (info->fix.visual) {
	case FB_VISUAL_PSEUDOCOLOR:
		ep93xxfb_palette_write(regno, red, green, blue, transp);
		break;
	case FB_VISUAL_TRUECOLOR:
		if (regno >= 16)
			return 1;

		red = CNVT_TOHW(red, info->var.red.length);
		green = CNVT_TOHW(green, info->var.green.length);
		blue = CNVT_TOHW(blue, info->var.blue.length);
		transp = CNVT_TOHW(transp, info->var.transp.length);
		((u32 *) (info->pseudo_palette))[regno] =
		  (red << info->var.red.offset) |
		  (green << info->var.green.offset) |
		  (blue << info->var.blue.offset) |
		  (transp << info->var.transp.offset);
		break;
	case FB_VISUAL_DIRECTCOLOR:
		red = CNVT_TOHW(red, 8);
		green = CNVT_TOHW(green, 8);
		blue = CNVT_TOHW(blue, 8);
		transp = CNVT_TOHW(transp, 8);
		break;
	}
#undef CNVT_TOHW

	return 0;
}

static int
ep93xxfb_blank(int blank_mode, struct fb_info *info)
{
	unsigned long attribs;

	attribs = inl(VIDEOATTRIBS) & ~VIDEOATTRIBS_SDSEL_MASK;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif

	if (blank_mode) {
		if (modedb[mode].off)
			(modedb[mode].off)(info);
		ep93xxfb_outl(attribs & ~(VIDEOATTRIBS_DATAEN |
					  VIDEOATTRIBS_SYNCEN |
					  VIDEOATTRIBS_PCLKEN |
					  VIDEOATTRIBS_EN),
		     VIDEOATTRIBS);
	} else {
		if (modedb[mode].clk_src == CLK_INTERNAL)
			attribs |= VIDEOATTRIBS_PCLKEN;
		ep93xxfb_outl(attribs | VIDEOATTRIBS_DATAEN
			      | VIDEOATTRIBS_SYNCEN
			      | VIDEOATTRIBS_EN,
			      VIDEOATTRIBS);
		if (modedb[mode].on)
			(modedb[mode].on)(info);
	}

	return 0;
}


/*
 * ep93xxfb_set_par():
 *
 */
static int
ep93xxfb_set_par(struct fb_info *info)
{
	
	switch (info->var.bits_per_pixel) {
	case 8:
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
		break;

	case 16:
	case 24:
	case 32:
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;

	default:
		return -EINVAL;
	}

	info->fix.line_length = (info->var.xres * info->var.bits_per_pixel) / 8;

	return 0;
}

static struct fb_ops ep93xxfb_ops = {
	.owner		= THIS_MODULE,
	.fb_setcolreg	= ep93xxfb_setcolreg,
	.fb_set_par     = ep93xxfb_set_par,
	.fb_blank	= ep93xxfb_blank,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor	= soft_cursor,
	.fb_ioctl       = ep93xxfb_ioctl,
};

static unsigned long
ep93xx_get_pll_frequency(unsigned long pll)
{
	unsigned long fb1, fb2, ipd, ps, freq;

	if (pll == 1)
		pll = inl(SYSCON_CLKSET1);
	else if (pll == 2)
		pll = inl(SYSCON_CLKSET2);
	else
		return 0;

	ps = (pll & SYSCON_CLKSET1_PLL1_PS_MASK) >>
		SYSCON_CLKSET1_PLL1_PS_SHIFT;
	fb1 = ((pll & SYSCON_CLKSET1_PLL1_X1FBD1_MASK) >>
	       SYSCON_CLKSET1_PLL1_X1FBD1_SHIFT) + 1;
	fb2 = ((pll & SYSCON_CLKSET1_PLL1_X2FBD2_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2FBD2_SHIFT) + 1;
	ipd = ((pll & SYSCON_CLKSET1_PLL1_X2IPD_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2IPD_SHIFT) + 1;

	freq = 225 * fb1 * fb2;

	freq = (((freq / ipd) * 65536) + (((freq % ipd) * 65536) / ipd)) >> ps;

	return freq;
}

static int
ep93xx_set_video_div(unsigned long freq)
{
	unsigned long pdiv = 0, div = 0, psel = 0, esel = 0, f, i, j, k, l;
	long err, e;

	err = LONG_MAX;

	for (i = 0; i < 3 && err; i++) {
		if (i == 0)
			f = 14745600 * 2;
		else if (i == 1)
			f = ep93xx_get_pll_frequency(1) * 2;
		else
			f = ep93xx_get_pll_frequency(2) * 2;

		for (j = 4; j <= 6 && err; j++) {
			l = f / (freq * j);
			for (k = l + 1; k >= l && err; k--) {
				if (k < 2)
					break;
				e = (long)(f / (j * k)) - (long)freq;
				if (abs(e) < abs(err)) {
					pdiv = j - 3;
					div = k;
					psel = (i == 2) ? 1 : 0;
					esel = (i == 0) ? 0 : 1;
					err = e;
				}
			}
		}
	}

	if (err == LONG_MAX)
		return -1;

	outl(0xaa, SYSCON_SWLOCK);
	outl(SYSCON_VIDDIV_VENA | (esel ? SYSCON_VIDDIV_ESEL : 0) |
	     (psel ? SYSCON_VIDDIV_PSEL : 0) |
	     (pdiv << SYSCON_VIDDIV_PDIV_SHIFT) |
	     (div << SYSCON_VIDDIV_VDIV_SHIFT), SYSCON_VIDDIV);

	return freq + err;
}

static int
ep93xxfb_configure(struct fb_info *info)
{
	unsigned long htotal, hres, hfp, hsync, hbp;
	unsigned long vtotal, vres, vfp, vsync, vbp;
	unsigned long fb_size, clk, attribs;
	void *fb_log;

	//outl(inl(SDRAMGLOBALCFG) | 0x80, SDRAMGLOBALCFG);

	fb_size = modedb[mode].hres * modedb[mode].vres * (bpp / 8);
	fb_log = dma_alloc_writecombine(NULL, fb_size, &fb_phys, GFP_KERNEL);
	if (!fb_log)
		return -ENOMEM;

	vres = modedb[mode].vres;
	vfp = modedb[mode].vfp;
	vsync = modedb[mode].vsync;
	vbp = modedb[mode].vbp;
	vtotal = vres + vfp + vsync + vbp;

	hres = modedb[mode].hres;
	hfp = modedb[mode].hfp;
	hsync = modedb[mode].hsync;
	hbp = modedb[mode].hbp;
	htotal = hres + hfp + hsync + hbp;

	if (modedb[mode].clk_src == CLK_INTERNAL) {
		outl(0xaa, SYSCON_SWLOCK);
		outl(inl(SYSCON_DEVCFG) & ~SYSCON_DEVCFG_EXVC, SYSCON_DEVCFG);
		clk = ep93xx_set_video_div(htotal * vtotal *
					   modedb[mode].refresh +
					   (htotal * vtotal *
					    modedb[mode].mrefresh + 500)
					   / 1000);
		if (clk == -1) {
			dma_free_writecombine(NULL, fb_size, fb_log, fb_phys);
			return -ENODEV;
		}

		info->var.pixclock = 1000000000 / (clk / 1000);
	} else {
		outl(0xaa, SYSCON_SWLOCK);
		outl(0, SYSCON_VIDDIV);
		outl(0xaa, SYSCON_SWLOCK);
		outl(inl(SYSCON_DEVCFG) | SYSCON_DEVCFG_EXVC, SYSCON_DEVCFG);
		info->var.pixclock = modedb[mode].refresh;
	}

	info->var.xres = hres;
	info->var.yres = vres;
	info->var.xres_virtual = hres;
	info->var.yres_virtual = vres;
	info->var.bits_per_pixel = bpp;
	if (bpp == 8) {
		info->var.red.length = 8;
		info->var.green.length = 8;
		info->var.blue.length = 8;
	} else if ( bpp == 16) {
		info->var.red.offset = 11;
		info->var.red.length = 5;
		info->var.green.offset = 5;
		info->var.green.length = 6;
		info->var.blue.offset = 0;
		info->var.blue.length = 5;
	    
	} else if ( bpp == 24 || bpp == 32 ) {
		info->var.red.offset = 16;
		info->var.red.length = 8;
		info->var.green.offset = 8;
		info->var.green.length = 8;
		info->var.blue.offset = 0;
		info->var.blue.length = 8;
	    
	}
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.height = modedb[mode].height;
	info->var.width = modedb[mode].width;
	info->var.left_margin = hbp;
	info->var.right_margin = hfp;
	info->var.upper_margin = vbp;
	info->var.lower_margin = vfp;
	info->var.hsync_len = hsync;
	info->var.vsync_len = vsync;
	if (modedb[mode].pol_hsync == POL_HIGH)
		info->var.sync |= FB_SYNC_HOR_HIGH_ACT;
	if (modedb[mode].pol_vsync == POL_HIGH)
		info->var.sync |= FB_SYNC_VERT_HIGH_ACT;
	info->var.vmode = FB_VMODE_NONINTERLACED;

	info->fix.smem_start = fb_phys;
	info->fix.smem_len = fb_size;
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	if (bpp == 8)
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
	else if (bpp == 16 || bpp == 24 || bpp == 32)
	       	info->fix.visual = FB_VISUAL_TRUECOLOR;

	info->fix.line_length = hres * (bpp / 8);
	info->screen_base = fb_log;

	ep93xxfb_outl(0, VIDEOATTRIBS);

	ep93xxfb_outl(vtotal - 1, VLINESTOTAL);
	ep93xxfb_outl((vtotal - vfp - 1) + ((vtotal - vfp - vsync - 1) << 16),
		      VSYNCSTRTSTOP);
	ep93xxfb_outl((vres - 1) + ((vtotal - 1) << 16), VACTIVESTRTSTOP);
	ep93xxfb_outl((vres - 1) + ((vtotal - 1) << 16), VBLANKSTRTSTOP);
	ep93xxfb_outl((vtotal - 1) + ((vtotal - 1) << 16), VCLKSTRTSTOP);

	ep93xxfb_outl(htotal - 1, HCLKSTOTAL);
	ep93xxfb_outl((htotal - 1) + ((htotal - hsync - 1) << 16),
		      HSYNCSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      HBLANKSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      HACTIVESTRTSTOP);
	ep93xxfb_outl((htotal - 1) + ((htotal - 1) << 16), HCLKSTRTSTOP);

	ep93xxfb_outl(0, LINECARRY);

	outl(fb_phys, VIDSCRNPAGE);
	outl(vres, SCRNLINES);
	outl(((hres * bpp) / 32) - 1, LINELENGTH);
	outl((hres * bpp) / 32, VLINESTEP);

	if (bpp == 8)
		outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_8BPP |
		     PIXELMODE_C_LUT, PIXELMODE);
	else if (bpp == 16)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_16BPP |
		     PIXELMODE_C_565, PIXELMODE);
	else if (bpp == 24)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_24BPP |
		     PIXELMODE_C_888, PIXELMODE);
	else if (bpp == 32)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_32BPP |
		     PIXELMODE_C_888, PIXELMODE);
	//outl(0x10, FIFOLEVEL);

	SysconSetLocked(SYSCON_DEVCFG,
			inl(SYSCON_DEVCFG) | SYSCON_DEVCFG_RasOnP3);

	attribs = 0;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
	if (modedb[mode].clk_edge == EDGE_RISING)
		attribs |= VIDEOATTRIBS_INVCLK;
	if (modedb[mode].pol_blank == POL_HIGH)
		attribs |= VIDEOATTRIBS_BLKPOL;
	if (modedb[mode].pol_hsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_HSPOL;
	if (modedb[mode].pol_vsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_VCPOL;
	ep93xxfb_outl(attribs, VIDEOATTRIBS);

	if (modedb[mode].configure)
		(modedb[mode].configure)(info);

	ep93xxfb_blank(0, info);

	return 0;
}

extern int ep93xxfb_setup(char *options);
static int
ep93xxfb_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res;
	struct fb_info *info=NULL;
	int ret;
#ifndef MODULE
	char *option = NULL;

	if (fb_get_options(devicename, &option))
		return -ENODEV;
	ep93xxfb_setup(option);
#endif

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENODEV;

	if (!request_mem_region(res->start, res->end - res->start + 1,
				devicename))
		return -EBUSY;

	if (bpp == 8)
		info = framebuffer_alloc(sizeof(u32) * 256, &pdev->dev);
	else if (bpp == 16 || bpp == 24 || bpp == 32)
		info = framebuffer_alloc(sizeof(u32) * 16, &pdev->dev);
	else {
		ret = -EINVAL;
		goto memregion;
	}

	info->flags = FBINFO_DEFAULT;
	strncpy(info->fix.id, devicename, sizeof(info->fix.id));
	info->fix.mmio_start = res->start;
	info->fix.mmio_len = res->end - res->start + 1;
	info->fbops = &ep93xxfb_ops;
	info->pseudo_palette = info->par;
	info->state = FBINFO_STATE_RUNNING;

	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		ret = -ENOMEM;
		goto framebuffer;
	}

	if ((ret = ep93xxfb_configure(info)) < 0)
		goto colormap;

	if (register_framebuffer(info) < 0) {
		printk(KERN_ERR "Unable to register EP93xx frame buffer\n");
		ret = -EINVAL;
		goto dma;
	}

	dev_set_drvdata(dev, info);

	printk(KERN_INFO "fb%d: EP93xx frame buffer at %dx%dx%d\n", info->node,
	       info->var.xres, info->var.yres, info->var.bits_per_pixel);

	return 0;

dma:
	dma_free_writecombine(NULL, info->fix.smem_len,
			      info->screen_base, info->fix.smem_start);
colormap:
	fb_dealloc_cmap(&info->cmap);
framebuffer:
	framebuffer_release(info);
memregion:
	release_mem_region(res->start, res->end - res->start + 1);

	return ret;
}

static int __devexit
ep93xxfb_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct fb_info *info = dev_get_drvdata(dev);
	struct resource *res;

	dev_set_drvdata(dev, NULL);

	unregister_framebuffer(info);

	dma_free_writecombine(NULL, info->fix.smem_len,
			      info->screen_base, info->fix.smem_start);
	fb_dealloc_cmap(&info->cmap);
	framebuffer_release(info);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(res->start, res->end - res->start + 1);

	return 0;
}

static  int
ep93xxfb_ioctl(struct inode *inode, struct file *file,
	       unsigned int cmd, unsigned long arg, struct fb_info *info)
{
	struct ep93xx_cursor cursor;
#ifdef CONFIG_EP93XX_GRAPHICS
	struct ep93xx_blit blit;
	struct ep93xx_fill fill;
	struct ep93xx_line line;
#endif
	unsigned long caps;

	switch (cmd) {
	case FBIO_EP93XX_GET_CAPS:
		caps = EP93XX_CAP_CURSOR;
#ifdef CONFIG_EP93XX_GRAPHICS
		caps |= EP93XX_CAP_LINE | EP93XX_CAP_FILL | EP93XX_CAP_BLIT;
#endif
		copy_to_user((void *)arg, &caps, sizeof(unsigned long));
		return 0;

	case FBIO_EP93XX_CURSOR:
		copy_from_user(&cursor, (void *)arg,
			       sizeof(struct ep93xx_cursor));
		return ep93xxfb_cursor(info,&cursor); 

#ifdef CONFIG_EP93XX_GRAPHICS
        case FBIO_EP93XX_LINE:
		copy_from_user(&line, (void *)arg, sizeof(struct ep93xx_line));
		return ep93xxfb_line(info,&line);

        case FBIO_EP93XX_FILL:
		copy_from_user(&fill, (void *)arg, sizeof(struct ep93xx_fill));
		return ep93xxfb_fill(info,&fill);

	case FBIO_EP93XX_BLIT:
		copy_from_user(&blit, (void *)arg, sizeof(struct ep93xx_blit));
		return ep93xxfb_blit(info,&blit);
#endif

	default:
		return -EFAULT;
	}
}

static struct device_driver ep93xxfb_driver = {
	.name		= devicename,
	.bus		= &platform_bus_type,
	.probe		= ep93xxfb_drv_probe,
	.remove		= __devexit_p(ep93xxfb_drv_remove),
};

int __init
ep93xxfb_init(void)
{
#ifdef MODULE
	int i, j;

	if (modestr) {
		for (i = 0; i < (sizeof(modedb) / sizeof(modedb[0])); i++) {
			for (j = 0; modedb[i].name[j]; j++)
				if (strcmp(modestr, modedb[i].name[j]) == 0)
					break;
			if (modedb[i].name[j])
				break;
		}
		if (i != (sizeof(modedb) / sizeof(modedb[0])))
			mode = i;
		else
			printk(KERN_ERR devicename ": unknown mode: %s\n",
			       modestr);
	}
#endif

#ifdef CONFIG_EP93XX_GRAPHICS
        outl(0x00000000, BLOCKCTRL);
        request_irq(IRQ_GRAPHICS, ep93xxfb_irq_handler, SA_INTERRUPT,
                    "graphics",NULL);
#endif

	return driver_register(&ep93xxfb_driver);
}

static void __exit
ep93xxfb_cleanup(void)
{
	return driver_unregister(&ep93xxfb_driver);
}

int
ep93xxfb_setup(char *options)
{
	char *opt;
	int i, j;

	if (!options || !*options)
		return 0;

	while ((opt = strsep(&options, ",")) != NULL) {
		if (!*opt)
			continue;
		for (i = 0; i < (sizeof(modedb) / sizeof(modedb[0])); i++) {
			for (j = 0; modedb[i].name[j]; j++)
				if (strcmp(opt, modedb[i].name[j]) == 0)
					break;
			if (modedb[i].name[j])
				break;
		}
		if (i != (sizeof(modedb) / sizeof(modedb[0]))) {
			mode = i;
			continue;
		}
		if (strcmp(opt, "8bpp") == 0) {
			bpp = 8;
			bytes_per_pixel = 1;
			pixel_mask = 3;
			pixel_shift = 2;
			pixel_format = 0x00040000;
			continue;
		} else if (strcmp(opt, "16bpp") == 0) {
			bpp = 16;
			bytes_per_pixel = 2;
			pixel_mask = 1;
			pixel_shift = 1;
			pixel_format = 0x00080000;
			continue;
		} else if (strcmp(opt, "24bpp") == 0) {
			bpp = 24;
			bytes_per_pixel = 3;
			pixel_mask = 0;
			pixel_shift = 0;
			pixel_format = 0x00100000;
			continue;
		} else if (strcmp(opt, "32bpp") == 0) {
			bpp = 32;
			bytes_per_pixel = 3;
			pixel_mask = 0;
			pixel_shift = 0;
			pixel_format = 0x00100000;
			continue;
		}

		printk (KERN_ERR devicename ": unknown parameter: %s\n", opt);
	}

	return 0;
}

module_init(ep93xxfb_init);
module_exit(ep93xxfb_cleanup);

MODULE_PARM(modestr, "s");
MODULE_PARM(bpp, "i");

MODULE_AUTHOR("Cirrus Logic");
MODULE_DESCRIPTION("EP93xx frame buffer driver");
MODULE_LICENSE("GPL");
